#include "gibbsMedLDA.h"
#include "stats.h"
#include "cokus.h"
#include "utils.h"
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <vector>
#include <map>
#include <time.h>
using namespace std;

void GibbsMedLDA::set_default_values()
{
	m_ptrnstats			= NULL;
	m_ptststats			= NULL;

	m_nK				= 10; // number of topics, K
	m_nV				= 0;
	m_nLabelNum			= 0; // number of classes, |C|
	m_dAlpha			= 1; 
	m_dBeta				= 1;
	m_dNuSqInv          = 1;
	m_dC				= 1;
	m_dB				= 0;
	m_nBurnIn           = 100;
	m_nMaxSamples       = 100;

	m_dEta				= NULL; 
	m_dLambdaInv        = NULL;
	m_dCov              = NULL;
	m_dCovInv           = NULL;
	m_dLowerTriangle    = NULL;
	m_dPr				= NULL; 
	m_mrg				= NULL;
	m_dPhi				= NULL;
	m_dTheta			= NULL;
	m_dEtabk_           = NULL;
	m_dSampleEta        = NULL;
	m_pIGsampler        = NULL;
}

GibbsMedLDA::~GibbsMedLDA()
{
	if (m_dEta) delete[] m_dEta;
	m_dEta = NULL;

	if ( m_dCov ) {
		for ( int i=0; i<m_nK; i++ ) {
			delete[] m_dCov[i];
			delete[] m_dCovInv[i];
			delete[] m_dLowerTriangle[i];
		}
		delete[] m_dCov;
		delete[] m_dCovInv;
		delete[] m_dLowerTriangle;
	}

	if ( m_dEtabk_ ) delete[] m_dEtabk_;
	if ( m_dSampleEta  ) {
		for ( int i=0; i<m_nMaxSamples; i++ ) {
			delete[] m_dSampleEta[i];
		}
		delete[] m_dSampleEta;
	}

	if (m_dPr) delete[] m_dPr;
	m_dPr = NULL;

	if (m_mrg) delete[] m_mrg;
	m_mrg = NULL;

	if (m_dPhi) {
		for(int i = 0; i < m_nK; ++i)
			delete[] m_dPhi[i];
		delete[] m_dPhi;
	}
	m_dPhi = NULL;

	if (m_dTheta) delete[] m_dTheta;
	m_dTheta = NULL;

	if (m_ptrnstats) delete m_ptrnstats;
	m_ptrnstats = NULL;

	if (m_ptststats) delete m_ptststats;
	m_ptststats	= NULL;

	if ( m_pMVGaussian ) delete m_pMVGaussian;
}

void GibbsMedLDA::init_est(int refLabel)
{
	Corpus* corpus = new Corpus();
	corpus->read_data_binary(m_params->train_filename, refLabel);
	init_est(corpus);
}

void GibbsMedLDA::init_est(Corpus* corpus)
{
	m_nK 			= m_params->NTOPICS;
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 
	m_nBurnIn       = m_params->BURN_IN;
	m_nMaxSamples   = m_params->MAX_SAMPLES;

	m_ptrnstats = new Stats();
	m_ptrnstats->new_stats(corpus, m_params);
	m_nV = m_ptrnstats->nV;

	const int m_nD = m_ptrnstats->nD;
	int **m_nw		= m_ptrnstats->nw;
	int *m_nwsum	= m_ptrnstats->nwsum;
	int **m_nd		= m_ptrnstats->nd;

	if (!m_dEta) m_dEta = new double[m_nK];
	memset(m_dEta, 0, sizeof(double) * m_nK);
	if (!m_dCov) {
		m_dCov = new double*[m_nK];
		m_dCovInv = new double*[m_nK];
		m_dLowerTriangle = new double*[m_nK];
		for ( int k=0; k<m_nK; k++ ) {
			m_dCov[k] = new double[m_nK];
			m_dCovInv[k] = new double[m_nK];
			m_dLowerTriangle[k] = new double[m_nK];
			memset(m_dCov[k], 0, sizeof(double) * m_nK);
			memset(m_dCovInv[k], 0, sizeof(double) * m_nK);
			memset(m_dLowerTriangle[k], 0, sizeof(double) * m_nK);
		}
	}
	m_dSampleEta = new double*[m_nMaxSamples];
	for ( int i=0; i<m_nMaxSamples; i++ ) {
		m_dSampleEta[i] = new double[ m_nK ];
		memset(m_dSampleEta[i], 0, sizeof(double) * m_nK);
	}

	if (!m_dLambdaInv) m_dLambdaInv = new double[corpus->num_docs];
	memset(m_dLambdaInv, 1, sizeof(double)*corpus->num_docs);
	if (!m_dEtabk_) m_dEtabk_ = new double[m_nK];
	memset(m_dEtabk_, 0, sizeof(double)*m_nK);

	if (!m_dPr) m_dPr = new double[m_nK];
	if (!m_mrg) m_mrg = new double[m_nK];

	int N, topic;
	int* zptr = NULL;
	for(int d = 0; d < m_nD; ++d) {
		Document& dc = corpus->docs[d];
		N = dc.length; 
		zptr = m_ptrnstats->z[d];
		for(int n = 0; n < N; ++n) {
			topic = (int)(myrand() * m_nK);	
			*zptr++ = topic;
			m_nw[topic][dc.words[n]] += 1;
			m_nwsum[topic] += 1;
			m_nd[d][topic] += 1;
		}
	}
	m_dTrainTime_Lambda = 0;
	m_dTrainTime_Eta = 0;
}

void GibbsMedLDA::init_estc(const char* prev_state_file)
{
	m_nK 			= m_params->NTOPICS; 
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 

	if (!m_ptrnstats) load_model(prev_state_file);

	if (!m_dPr) m_dPr = new double[m_nK];
	if (!m_mrg) m_mrg = new double[m_nK];

	m_dTrainTime_Lambda = 0;
	m_dTrainTime_Eta = 0;
}

void GibbsMedLDA::init_inf(Corpus* corpus)
{
	m_nK			= m_params->NTOPICS;
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 
	m_nBurnIn       = m_params->BURN_IN;
	m_nMaxSamples   = m_params->MAX_SAMPLES;

	if (m_ptststats) delete m_ptststats;
	m_ptststats = new Stats();
	m_ptststats->new_stats(corpus, m_params);
	if (m_ptststats->nV > m_nV) 
		m_nV = m_ptststats->nV;

	if(!m_dPr) m_dPr = new double[m_nK];

	const int infD	= m_ptststats->nD;
	int **infnw		= m_ptststats->nw;
	int *infnwsum	= m_ptststats->nwsum;
	int *zptr 		= NULL;
	int *infnd 		= NULL;
	int topic, term;
	for (int d = 0; d < infD; ++d) {
		Document& dc 	= corpus->docs[d];
		zptr				= m_ptststats->z[d];
		infnd				= m_ptststats->nd[d];
		for( int n = 0; n < dc.length; ++n) {
			term = dc.words[n];
			topic = (int)(myrand() * m_nK);
			*zptr++ = topic;
			infnw[topic][term] 	+= 1;
			infnwsum[topic]		+= 1;
			infnd[topic]			+= 1;
		}
	}
}

void GibbsMedLDA::init_inf(const char* model_file)
{
	load_model(model_file);

	Corpus* corpus = new Corpus();
#ifdef _REMOVE_OOV_
	corpus->read_data_oov(m_params->test_filename, m_nV);
#else
	corpus->read_data_binary(m_params->test_filename);
#endif
	init_inf(corpus);
}

void GibbsMedLDA::train()
{
	double dStartTime = get_runtime();
	Corpus *pTrC = m_ptrnstats->corpus;
	m_pIGsampler = new InverseGaussian(1, 1);

	// burn in stage
	printf(">> burn-in stage (iter %d): ", m_nBurnIn);
	for ( int it=0; it<m_nBurnIn; it++ ) {
		printf(".");
		double dtime1 = get_runtime();
		drawLambda();
		double dtime2 = get_runtime();
		m_dTrainTime_Lambda += (dtime2 - dtime1) / 100;

		drawEta();
		dtime1 = get_runtime();
		m_dTrainTime_Eta += (dtime1 - dtime2) / 100;
		drawZ_single_iter();
	}

	// sample stage (keep the samples)
	printf("\n>> burn-in done \n");
	printf(">> draw samples (iter %d): ", m_nMaxSamples);
	int iter = 0;
	for ( ; iter<m_nMaxSamples; iter++ ) {
		printf(".");
		double dtime1 = get_runtime();
		drawLambda();
		double dtime2 = get_runtime();
		m_dTrainTime_Lambda += (dtime2 - dtime1) / 100;

		drawEta();
		dtime1 = get_runtime();
		m_dTrainTime_Eta += (dtime1 - dtime2) / 100;
		for ( int k=0; k<m_nK; k++ ) {
			m_dSampleEta[iter][k] = m_dEta[k];
		}
		drawZ_single_iter();
	}
	printf(">> draw samples done \n");

	// the average is the final weight
	memset(m_dEta, 0, sizeof(double)*m_nK);
	for ( int n=0; n<m_nMaxSamples; n++ ) {
		addvec(m_dEta, m_dSampleEta[n], m_nK, 1.0/m_nMaxSamples);
	}
	double dEndTime = get_runtime();
	m_dTrainTime = (dEndTime - dStartTime) / 100;

	char buf[1024];
	sprintf(buf, "%s/model.txt", m_params->dir);
	save_model(buf);

	sprintf(buf, "%s/lambda.txt", m_params->dir);
	save_lambda(buf);

	sprintf(buf, "%s/time.txt", m_params->dir);
	FILE *ftime = fopen(buf, "w");
	fprintf(ftime, "K: %d\n", m_nK);
	fprintf(ftime, "Total training time: %lf\n", m_dTrainTime);
	fclose(ftime);

	//printf("Final E-step\n");
	//double lhood = drawZ(iter);
	//printf("Final E-step Likelihood: %5.10f\n", lhood);
	//printf("Final E-step finished!\n");

	sprintf(buf, "%s/topwords.txt", m_params->dir);
	m_ptrnstats->save_topwords(buf, m_params);

	m_dTrainAcc = 0;
	for (int d = 0; d < m_ptrnstats->nD; ++d) {
		Document& dc = pTrC->docs[d];
		dc.predlabel = predict(m_ptrnstats, d);	
		if (dc.gndlabel == dc.predlabel) ++m_dTrainAcc;
	}
	m_dTrainAcc /= m_ptrnstats->nD;
	printf("\n\n Double count accuracy: %.5f\n\n", m_dTrainAcc);
	sprintf(buf, "%s/double_count_accuracy.txt", m_params->dir);
	FILE* dfile = fopen(buf, "w");
	fprintf(dfile, "Double count accuracy: %.5f\n", m_dTrainAcc);
	fclose(dfile);

	// free memory
	delete m_pIGsampler;
}

void GibbsMedLDA::run_em()
{
	m_pIGsampler = /*new InverseGaussian*[m_ptrnstats->corpus->num_docs];
	for ( int d=0; d<m_ptrnstats->corpus->num_docs; d++ ) {
		m_pIGsampler[d] = */new InverseGaussian(1, 1);
	//}

	char buf[512];
	double lhood = 1;
	double lhood_old = 0;
	double converged = 1;
	int iter = 0;

	double svmTime = 0.0;
	time_t rawtime;
	struct tm *timeinfo = NULL;
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	printf("K: %d >> Training start time: %s\n", m_nK, asctime(timeinfo));
	fflush(stdout);

	clock_t totalStart = clock();
	if (m_params->EM_CONVERGED >= 0)
	{
		sprintf(buf, "%s/train_likelihood.txt", m_params->dir);
		FILE* flhood = fopen(buf, "w");
		sprintf(buf, "%s/ulda.model", m_params->dir);
		while((converged > m_params->EM_CONVERGED) && (iter < m_params->EM_MAX_ITER) || (m_params->EM_MAX_ITER < 0))
		{
			clock_t start = clock();
			drawLambda();
			drawEta();
			clock_t end = clock();
			svmTime += double(end - start)/CLOCKS_PER_SEC;

			printf("----- %d-th EM ------\n", iter);
			lhood = drawZ(iter);
			fflush(stdout);

			//if (0 == iter) save_model(buf);
			converged = lhood_old - lhood;
			converged = fabs(converged/lhood_old);
			lhood_old = lhood;
			printf("----------------------------\n");
			printf("|  likelihood: %.10f; convergence: %.10f\n", lhood, converged);
			fprintf(flhood, "likelihood: %.10f; convergence: %.10f", lhood, converged);
			printf("----------------------------\n");
			printf("--- %d-th EM finished ----\n", iter);
			iter++;
		}
		fclose(flhood);
	} else {
		for(iter = 0; iter < m_params->EM_MAX_ITER; ++iter) {
			printf("----- %d-th EM ------\n", iter);
			drawZ(iter);
			drawEta();
			drawLambda();
			printf("--- %d-th EM finished ----\n", iter);
		}
	}
	clock_t totalEnd = clock();
	m_dTrainTime = (double)(totalEnd - totalStart)/CLOCKS_PER_SEC;

	sprintf(buf, "%s/time.txt", m_params->dir);
	FILE *ftime = fopen(buf, "w");
	fprintf(ftime, "K: %d\n", m_nK);
	fprintf(ftime, "Training start time: %s\n", asctime(timeinfo)); 
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	fprintf(ftime, "Training end time: %s\n", asctime(timeinfo)); 
	fprintf(ftime, "Total training time: %lf\n", m_dTrainTime);
	fprintf(ftime, "Lambda-Eta time: %lf\n", svmTime); 
	fclose(ftime);
	printf("K: %d >> Training end time: %s\n", m_nK, asctime(timeinfo)); 
	printf("K: %d >> Lambda-Eta time: %lf\n", m_nK, svmTime); 
	fflush(stdout);

	printf("Final E-step\n");
	lhood = drawZ(iter);
	printf("Final E-step Likelihood: %5.10f\n", lhood);
	printf("Final E-step finished!\n");

	sprintf(buf, "%s/topwords.txt", m_params->dir);
	m_ptrnstats->save_topwords(buf, m_params);

	int nAcc = 0;
	int label = 0;
	for (int d = 0; d < m_ptrnstats->nD; ++d) {
		label = predict(m_ptrnstats, d);	
		Document& dc = m_ptrnstats->corpus->docs[d];
		dc.predlabel = label;
		if (dc.gndlabel == label) ++nAcc;
	}
	printf("\n\n Double count accuracy: %.5f\n\n", (double)nAcc/m_ptrnstats->nD);
	sprintf(buf, "%s/double_count_accuracy.txt", m_params->dir);
	FILE* dfile = fopen(buf, "w");
	fprintf(dfile, "Double count accuracy: %.5f\n", (double)nAcc/m_ptrnstats->nD);
	fclose(dfile);

	//for ( int d=0; d<m_ptrnstats->corpus->num_docs; d++ ) {
	//	delete m_pIGsampler[d];
	//}
	//delete[] m_pIGsampler;
	delete m_pIGsampler;
}

void GibbsMedLDA::drawLambda()
{
	const int m_nD = m_ptrnstats->nD;
	double dMu;
	double cfactor = m_dC * m_dC;

	for ( int d=0; d<m_nD; d++ ) {
		Document& dc = m_ptrnstats->corpus->docs[d];

		dMu = m_dC / fabs( m_dDeltaEll - dc.gndlabel * discFunc( d, dc.length ) );

		m_pIGsampler->reset( dMu, cfactor );
		m_dLambdaInv[d] = m_pIGsampler->sample();
	}
}

double GibbsMedLDA::discFunc(const int &d, const int &dlen) 
{
	double dval = 0;
	int *zptr = m_ptrnstats->nd[d];

	for ( int k=0; k<m_nK; k++ ) {
		dval += m_dEta[k] * zptr[k];
	}
	dval /= dlen;

	return dval;
}

void GibbsMedLDA::drawEta()
{
	double *dCovInvPtr = NULL;
	double dFactor, dLocalFactor;

	// initialization
	memset(m_dEta, 0, sizeof(double)*m_nK);
	for ( int k=0; k<m_nK; k++ ) {
		memset(m_dCovInv[k], 0, sizeof(double)*m_nK);
		m_dCovInv[k][k] = m_dNuSqInv;
	}

	Corpus *pTrC = m_ptrnstats->corpus;
	Document *pDoc = NULL;
	for ( int d=0; d<pTrC->num_docs; d++ ) {
		pDoc = &(pTrC->docs[d]);
		int *cptr = m_ptrnstats->nd[d];

		// covariance
		dFactor = m_dLambdaInv[d] / (pDoc->length * pDoc->length);
		for ( int i=0; i<m_nK; i++ ) {
			if ( cptr[i] > 0 ) {
				dCovInvPtr = m_dCovInv[i];
				dLocalFactor = dFactor * cptr[i];
				for ( int j=0; j<m_nK; j++ ) {
					dCovInvPtr[j] += dLocalFactor * cptr[j];
				}
			}
		}

		// mean
		dFactor = pDoc->gndlabel * (m_dDeltaEll*m_dLambdaInv[d] + m_dC) / pDoc->length;
		for ( int i=0; i<m_nK; i++ ) {
			if ( cptr[i] > 0 ) 
				m_dEta[i] += cptr[i] * dFactor;
		}
	}

	///* Strategy #1: use the posterior mean as the 'sample' */
	//// matrix inverse
	//inverse( m_dCovInv, m_dCov, m_nK );	

	//for ( int k=0; k<m_nK; k++ ) {
	//	m_dEtabk_[k] = dotprod( m_dCov[k], m_dEta, m_nK );
	//}
	//for ( int k=0; k<m_nK; k++ ) {
	//	m_dEta[k] = m_dEtabk_[k];
	//}

	/* Strategy #2: draw random samples from multivariate Gaussian */
	inverse_cholydec(m_dCovInv, m_dCov, m_dLowerTriangle, m_nK);
	for ( int k=0; k<m_nK; k++ ) {
		m_dEtabk_[k] = dotprod( m_dCov[k], m_dEta, m_nK );
	}
	if ( m_pMVGaussian == NULL ) m_pMVGaussian = new MVGaussian();
	m_pMVGaussian->nextMVGaussianWithCholesky(m_dEtabk_, m_dLowerTriangle, m_dEta, m_nK);
}

double GibbsMedLDA::drawZ_single_iter(bool bLhood)
{
	const int m_nD = m_ptrnstats->nD;
	double delta = 0;
	int *zptr = NULL;
	int *ndptr = NULL;
	bool log_space_flag = false;
	double gamma, dfactor;
	int nlen;

	// sample topic assignment Z
	for(int d = 0; d < m_nD; ++d)
	{
		Document& dc = m_ptrnstats->corpus->docs[d];
		nlen         = dc.length;
		zptr         = m_ptrnstats->z[d]; 
		ndptr        = m_ptrnstats->nd[d];
		log_space_flag = compute_mrgterm_doc(d, dc.gndlabel, nlen, m_mrg);
		gamma = 1.0 / (double)nlen;
		dfactor = - gamma * (1 - gamma) * m_dLambdaInv[d];

		if (log_space_flag) {
			for(int n = 0; n < nlen; ++n) {
				delta = compute_delta(ndptr, zptr[n], nlen) * dfactor;
				zptr[n] = gibbs_sampling_logspace(m_ptrnstats, m_mrg, delta, d, n);
			}
		} else {
			for(int n = 0; n < nlen; ++n) {
				delta = compute_delta(ndptr, zptr[n], nlen) * dfactor;
				zptr[n]	= gibbs_sampling(m_ptrnstats, m_mrg, delta, d, n);
			}
		}
	}

	double lhood = 0;
	if ( bLhood ) lhood = compute_lhood(m_ptrnstats);

	return lhood;
}

double GibbsMedLDA::drawZ(int em_it)
{
	char model_file[500];
	double lhood = 0;
	double lhood_old = 0;
	double converged = 1;

	printf("gibbs_conv: %.10lf\tmax_iters:%d\n", m_params->GIBBS_CONVERGED, m_params->NITERS);
	if (m_params->GIBBS_CONVERGED >= 0)
	{
		int iter = 0;
		while((iter < MIN_GIBBS_ITERATION) || (converged > m_params->GIBBS_CONVERGED) && (iter < m_params->NITERS) || (m_params->NITERS < 0))
		{
			printf("Iteration %d ..., ", iter);

			lhood = drawZ_single_iter(true);

			if ((m_params->ITER_SAVE_STEP > 0) && (iter % m_params->ITER_SAVE_STEP == 0)) {
				sprintf(model_file, "%s/model-%d-%d", m_params->dir, em_it, iter);
				save_model(model_file);
			}
			//lhood = compute_lhood(m_ptrnstats);
			converged = (lhood_old - lhood);
			converged = fabs(converged/lhood_old);
			printf("\tlhood: %f\tdelta: %f\n", lhood, converged);
			lhood_old = lhood;
			++iter;
		}
		printf("Done! \tlhood: %f\tdelta: %f\n", lhood, converged);
	}
	else;

	return lhood;
}

double GibbsMedLDA::inference(double *tAvgEnt)
{
	char buf[512];
	sprintf(buf, "%s/inf_single_likelihood.txt", m_params->dir);

	int nAcc	= 0;

	double converged  = 1;
	double lhood = 0, lhood_old = 0;
	double lhood_all = 0;
	double avgEnt = 0, dEnt, val;
	int *zptr 			= NULL;
	int zLen, n, iter, topic, k;

	const int infD 	= m_ptststats->nD;
	const int infV		= m_ptststats->nV;
	Document* docs 	= m_ptststats->corpus->docs;
	int** infnw			= m_ptststats->nw;
	int* infnwsum		= m_ptststats->nwsum;

	if (m_dPhi) {
		for(k = 0; k < m_nK; ++k) {
			if (m_dPhi[k]) delete[] m_dPhi[k];
			m_dPhi[k] = NULL;
		}
		delete[] m_dPhi;
	}
	m_dPhi = NULL;

	if (m_params->GIBBS_CONVERGED >= 0) 
		compute_phi(m_ptrnstats);
	int *suff_topics = new int[m_nK];

	time_t rawtime;
	struct tm *timeinfo;
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	printf("K: %d >> Test start time: %s\n", m_nK, asctime(timeinfo));
	fflush(stdout);

	double dStartTime = get_runtime();
	printf("Prediction Inference ...\n");
	for(int d = 0; d < infD; ++d) {
		memset(infnwsum, 0, sizeof(int) * m_nK);
		for(k = 0; k < m_nK; ++k)
			memset(infnw[k], 0, sizeof(int)* infV);

		Document& dc = docs[d];
		zLen = dc.length; 
		zptr = m_ptststats->z[d];
		for(n = 0; n < zLen; ++n) {
			topic = zptr[n];
			infnw[topic][dc.words[n]] += 1;
			infnwsum[topic] += 1;
		}


		if (d % 1000 == 0) printf("Document %d\n", d);

		// burn-in stage
		iter = 0;
		lhood = 0;
		if (m_params->GIBBS_CONVERGED >= 0) {
			converged  = 1;
			lhood_old = 0;
		  	while((converged > m_params->GIBBS_CONVERGED) && (iter < m_params->NITERS) || (m_params->NITERS < 0)) {
		  		for(n = 0; n < zLen; ++n)
		  			zptr[n] = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
				compute_theta(m_ptststats, d);
		  		lhood = compute_lhood(&dc, m_dPhi, m_dTheta);
		  		converged = (lhood_old - lhood);
		  		converged = fabs(converged/lhood_old);
		  		lhood_old = lhood;
		  		++iter;
		  	}
		} else {
			for (iter = 0; iter < m_params->NITERS; ++iter)
			  	for(n = 0; n < zLen; ++n)
			  		zptr[n] = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
		}
		lhood_all += lhood;
		dc.lhood  = lhood;

		// predicting labels based on the sufficient statistics
		memset(suff_topics, 0, sizeof(int) * m_nK);
		for (iter = 0; iter < m_params->SAMPLING_LAG; ++iter) {
		  	for(n = 0; n < zLen; ++n) {
				topic = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
				zptr[n] = topic;
				suff_topics[topic] += 1;
			}
		}

		if (m_params->SAMPLING_LAG > 1){
			dEnt = 0;
			for(k = 0; k < m_nK; ++k) {
				val = (double)(suff_topics[k]/(double)m_params->SAMPLING_LAG + m_dAlpha)/(dc.length + m_nK * m_dAlpha);
				if (val > 0) dEnt -= val * log(val);
				m_dTheta[k] = val;
			}
			avgEnt += dEnt;
		} else if (m_params->SAMPLING_LAG == 1){
			double DV = (double)dc.length + m_nK * m_dAlpha;
			dEnt = 0;
			for(k = 0; k < m_nK; ++k) {
				val = ((double)suff_topics[k]+ m_dAlpha)/DV;
				dEnt -= val * log(val);
				m_dTheta[k] = val;
			}
			avgEnt += dEnt;
		} else ;

		predict(m_dTheta, dc);
		if (dc.predlabel == dc.gndlabel ) ++nAcc;
	}
	m_dTestTime = (get_runtime() - dStartTime) / 100;

	sprintf(buf, "%s/time.txt", m_params->dir);
	FILE *ftime = fopen(buf, "a");
	fprintf(ftime, "Test start time: %s\n", asctime(timeinfo));
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	fprintf(ftime, "Test end time: %s; Total: %.5f seconds \n", asctime(timeinfo), m_dTestTime);
	fclose(ftime);
	printf("K: %d >> Test end time: %s\n", m_nK, asctime(timeinfo));
	fflush(stdout);

	printf("Likilihood_ALL : %5.10f\n", lhood_all);
	printf("Prediction Inference FINISHED !\n");
	*tAvgEnt = avgEnt / infD;
	delete[] suff_topics;
	return (double) nAcc / infD;
}

int GibbsMedLDA::gibbs_sampling_logspace(Stats* ss, const double* mrg, const double &delta, const int d, const int n)
{
	Document& dc		= ss->corpus->docs[d];
	const int term		= dc.words[n];
	//const int y			= dc.gndlabel;
	//const int len		= dc.length;
	//int **m_z			= ss->z;
	int **m_nw			= ss->nw;
	int *m_nwsum		= ss->nwsum;
	int *ndptr			= ss->nd[d];

	int topic 			= ss->z[d][n];
	double Vbeta		= m_nV * m_dBeta;

	m_nw[topic][term]   -= 1;
	m_nwsum[topic]		-= 1;
	ndptr[topic]		-= 1;

	for(int k = 0; k < m_nK; ++k)
		m_dPr[k] = log((m_nw[k][term] + m_dBeta) / (m_nwsum[k] + Vbeta) * (ndptr[k] + m_dAlpha)) + (mrg[k] + delta*m_dEta[k]);
	double logsum = log_sum(m_dPr, m_nK);

	m_dPr[0] = exp(m_dPr[0] - logsum);
	for(int k = 1; k < m_nK; ++k)
		m_dPr[k] = m_dPr[k-1] + exp(m_dPr[k] - logsum); 

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_nw[topic][term]	+= 1;
	m_nwsum[topic]		+= 1;
	ndptr[topic]		+= 1;

	return topic;
}

int GibbsMedLDA::gibbs_sampling(Stats* ss, const double* mrg, const double &delta, const int d, const int n)
{
	Document& dc		= ss->corpus->docs[d];
	const int term		= dc.words[n];
	//const int y			= dc.gndlabel;
	//const int len		= dc.length;
	//int **m_z			= ss->z;
	int **m_nw			= ss->nw;
	int *m_nwsum		= ss->nwsum;
	int *ndptr			= ss->nd[d];

	int topic 			= ss->z[d][n];
	double Vbeta		= m_nV * m_dBeta;

	m_nw[topic][term]   -= 1;
	m_nwsum[topic]		-= 1;
	ndptr[topic]		-= 1;

	m_dPr[0] = (m_nw[0][term] + m_dBeta) / (m_nwsum[0] + Vbeta) * (ndptr[0] + m_dAlpha) * exp(mrg[0] + delta*m_dEta[0]);
	for(int k = 1; k < m_nK; ++k)
		m_dPr[k] = m_dPr[k-1] + (m_nw[k][term] + m_dBeta) / (m_nwsum[k] + Vbeta) * (ndptr[k] + m_dAlpha) * exp(mrg[k] + delta*m_dEta[k]); 

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_nw[topic][term]	+= 1;
	m_nwsum[topic]		+= 1;
	ndptr[topic]		+= 1;

	return topic;
}

bool GibbsMedLDA::compute_mrgterm_doc(const int d, const int y, const int dlen, double* mrg)
{
	double dval, eta, ov;
	double lambdaInv = m_dLambdaInv[d];
	double gamma = 1.0 / (double)(dlen);
	double gammaLambda = gamma * lambdaInv;
	double lambdaC = m_dC / lambdaInv;

	bool flag = false;
	for(int k = 0; k < m_nK; ++k) {
		eta = m_dEta[k];

		dval = gammaLambda * eta * ( y*(m_dDeltaEll + lambdaC) - 0.5*gamma*eta );

		if (dval > OVERFLOW_THREHOLD || dval < -OVERFLOW_THREHOLD) flag = true;
		mrg[k] = dval;
	}
	return flag;
}

// discriminant function value with word n removed.
double GibbsMedLDA::compute_delta(int *ndptr, int topic, const int &dlen)
{
	double dval = 0;
	if ( dlen > 1 ) {
		for (int k = 0; k < m_nK; ++k) {
			dval += m_dEta[k] * ndptr[k];
		}
		dval -= m_dEta[topic];
		dval /= (dlen - 1);
	}
	return dval;
}

void GibbsMedLDA::compute_theta(Stats* ss, int d)
{
	if (!m_dTheta)
		m_dTheta = new double[m_nK];	

	const double DIV = m_dAlpha * m_nK + ss->corpus->docs[d].length;
	int* nd_ptr = ss->nd[d];
	double* theta_ptr = m_dTheta;
	int k = m_nK;
	while(k--)
		*theta_ptr++ = double (*nd_ptr++ + m_dAlpha) / DIV; 
}

bool GibbsMedLDA::compute_phi(Stats* st)
{
	if ( st->nV > m_nV) {
		printf(" ERROR> V illeagle\n");	
		return false;
	}

	if(!m_dPhi) {
		m_dPhi = new double*[m_nK];
		for(int k = 0; k < m_nK; ++k)
			m_dPhi[k] = new double[m_nV];
	}

	const double Vbeta	= m_nV * m_dBeta;
	int t, nwsum;
	double* phi_ptr = NULL;
	int* nw_ptr = NULL;
	for(int k = 0; k < m_nK; ++k)
	{
		phi_ptr 	= m_dPhi[k];
		nw_ptr	= st->nw[k];
		const double div	= st->nwsum[k] + Vbeta;
		for (t = 0; t < st->nV; ++t)
			*phi_ptr++ = double (*nw_ptr++ + m_dBeta) / div; 
		for (; t < m_nV; ++t)
			*phi_ptr++ = m_dBeta / div;
	}
	return true;
}

double GibbsMedLDA::compute_lhood(Stats* st)
{
	double lhood = 0.0;
	double dlhood = 0, dlword, theta_k;
	int term, k, n;
	int *nd_ptr;
	bool log_space_flag = false;

	compute_phi(st);
	for(int d = 0; d < st->nD; ++d)
	{
		dlhood = 0;
		Document& dc = st->corpus->docs[d];	
		log_space_flag = compute_mrgterm_doc(d, dc.gndlabel, dc.length, m_mrg);
		compute_theta(st, d);
		for(k = 0; k < m_nK; ++k)
			dlhood += m_dTheta[k] * m_mrg[k] * dc.length;

		if (log_space_flag) {
			for(k = 0; k < m_nK; ++k)
				m_dTheta[k] = log(m_dTheta[k]) + m_mrg[k];
			for (n = 0; n < dc.length; ++n) {
				term = dc.words[n];	
				dlword = 0;
				for(k = 0; k < m_nK; ++k)
					dlword = log_sum(dlword, log(m_dPhi[k][term]) + m_dTheta[k]);
				dlhood -= dlword;
			}
		} else {
			for(k = 0; k < m_nK; ++k)
				m_dTheta[k] *= exp(m_mrg[k]);
			for (n = 0; n < dc.length; ++n) {
				dlword = 0;
				term = dc.words[n];	
				for(k = 0; k < m_nK; ++k)
					dlword += m_dPhi[k][term] * m_dTheta[k];
				dlhood -= log(dlword);
			}
		}
		lhood += dlhood;
	}
	return lhood;
}


double GibbsMedLDA::compute_lhood(Document* dc, double** phi, double* theta) {
	double dlhood = 0, dlword;
	int term;
	for (int n = 0; n < dc->length; ++n)
	{
		dlword = 0;
		term = dc->words[n];	
		for(int k = 0; k < m_nK; ++k)
			dlword += phi[k][term] * theta[k];
		dlhood -= log(dlword);
	}
	return dlhood;
}

int GibbsMedLDA::gibbs_sampling(Stats* trn_ss, Stats* tst_ss, const int d, const int n)
{
	double Vbeta		= m_dBeta * m_nV;

	Document& dc		= tst_ss->corpus->docs[d];
	const int term		= dc.words[n];
	int topic 			= tst_ss->z[d][n];
	int **m_infnw		= tst_ss->nw;
	int *m_infnwsum	= tst_ss->nwsum;
	int **m_infnd		= tst_ss->nd;

	int **m_nw			= trn_ss->nw;
	int *m_nwsum		= trn_ss->nwsum;
	int **m_nd			= trn_ss->nd;


	m_infnw[topic][term] -= 1;
	m_infnwsum[topic]		-= 1;
	m_infnd[d][topic]		-= 1;

	if ( term < trn_ss->nV)
	{
		m_dPr[0] = (m_nw[0][term] + m_infnw[0][term] + m_dBeta) / (m_nwsum[0] + m_infnwsum[0] + Vbeta) * (m_infnd[d][0] + m_dAlpha);
		for(int k = 1; k < m_nK; ++k)
			m_dPr[k] = m_dPr[k-1] + (m_nw[k][term] + m_infnw[k][term] + m_dBeta) / (m_nwsum[k] + m_infnwsum[k] + Vbeta) * (m_infnd[d][k] + m_dAlpha);
	}
	else 
	{
		m_dPr[0] = (m_infnw[0][term] + m_dBeta) / (m_infnwsum[0] + Vbeta) * (m_infnd[d][0] + m_dAlpha); 
		for(int k = 1; k < m_nK; ++k)
			m_dPr[k] = m_dPr[k-1] + (m_infnw[k][term] + m_dBeta) / (m_infnwsum[k] + Vbeta) * (m_infnd[d][k] + m_dAlpha);
	}

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_infnw[topic][term]	+= 1;
	m_infnwsum[topic]		+= 1;
	m_infnd[d][topic]		+= 1;

	return topic;
}

void GibbsMedLDA::predict(double *theta, Document &dc)
{
	double dScore = 0;										
	int k = m_nK;
	double *ptr = theta;
	double *eta_ptr = m_dEta;
	while(k--)
		dScore += (*eta_ptr++) * (*ptr++); 
	dScore -= m_dB;

	int pred_label = -1;		
	if ( dScore > 0 ) pred_label = 1;
 
	dc.predlabel = pred_label;
	dc.predscore = dScore;
}

int GibbsMedLDA::predict(int* topics, int doc_len, int num_iters)
{
	const double Kalpha = m_nK * m_dAlpha;
	double* eta_ptr = m_dEta;

	double dScore = 0;
	int k = m_nK;
	int *nd_ptr = topics;
	while(k--)
		dScore += (*eta_ptr++) * (double (*nd_ptr++) / double(num_iters) + m_dAlpha) / (doc_len + Kalpha); 
	dScore -= m_dB;

	int pred_label = -1;		
	if ( dScore > 0 ) pred_label = 1;

	return pred_label;
}

int GibbsMedLDA::predict(Stats* ss, const int d)
{
	const int len	= ss->corpus->docs[d].length;
	const double Kalpha = m_nK * m_dAlpha;

	double dScore = 0;										
	int k = m_nK;
	int *nd_ptr = ss->nd[d];
	double* eta_ptr = m_dEta;
	while(k--)
		dScore += (*eta_ptr++) * (*nd_ptr++ + m_dAlpha) / (len + Kalpha);
	dScore -= m_dB;
	
	int pred_label = -1;		
	if ( dScore > 0 ) pred_label = 1;
	return pred_label;
}

void GibbsMedLDA::outputLowDimData(Stats* ss, char *filename)
{
	double val;
	double Kalpha = m_nK * m_dAlpha; 
	FILE *fileptr = fopen(filename, "w");
	for ( int d = 0; d < m_ptrnstats->nD; d++ ) 
	{
		Document& dc = ss->corpus->docs[d];
		fprintf(fileptr, "%d %d", m_nK, dc.gndlabel);

		int* zz = ss->nd[d];
		for ( int k = 0; k < m_nK; k++ ) {
			val = (double(zz[k]) + m_dAlpha) / (double(dc.length) + Kalpha);
			fprintf(fileptr, " %d:%.10f", k, val);
		}
		fprintf(fileptr, "\n");
	}
	fclose(fileptr);
}

void GibbsMedLDA::save_results(double tAvgEnt)
{
	printf(">>> Saving results & evaluating performace\n");
	char buf[512];

	int nAcc = 0;
	int total = 0;
	double lhood = 0;
	for(int d = 0; d < m_ptststats->nD; ++d)
	{
		Document &dc = m_ptststats->corpus->docs[d];
		lhood	+= dc.lhood;
		total += dc.length;
		if (dc.gndlabel == dc.predlabel)
			++nAcc;
	}
	double dAcc = (double) nAcc / m_ptststats->nD;
	double perwordlhood = lhood / double(total);
	printf("Accuracy: %5.5f\n", dAcc);

	sprintf(buf, "%s/evl-performance.txt", m_params->dir);
	FILE* fileptr = fopen(buf, "w");
	fprintf(fileptr, "num docs: %d\n", m_ptststats->nD);
	fprintf(fileptr, "num terms: %d\n", m_nV);
	fprintf(fileptr, "num topics: %d\n", m_nK);
	fprintf(fileptr, "cost parameter l: %d\n", (int)m_dDeltaEll);
	fprintf(fileptr, "Alpha: %5.10f\n", (m_dAlpha*m_nK)); 
	//fprintf(fileptr, "Sample Lag: %d\n", m_params->SAMPLING_LAG);
	fprintf(fileptr, "C:	%5.10f\n\n", m_dC);
	fprintf(fileptr, "Accuracy: %5.10f\n", dAcc);
	fprintf(fileptr, "Likelihood: %5.10f\n", lhood);
	fprintf(fileptr, "PerWordLikelihood: %5.10f\n", perwordlhood);
	fprintf(fileptr, "testAverageEntropy: %5.10f\n", tAvgEnt);
	fclose(fileptr);

	sprintf(buf, "%s/pred_wordassigned.txt", m_params->dir);
	m_ptststats->save_wordassigned(buf);
	sprintf(buf, "%s/pred_results.txt", m_params->dir);
	m_ptststats->save_label_only(buf);

	sprintf(buf, "%s/gml_%dtopic_test.txt", m_params->dir, m_params->NTOPICS);
	FILE* fout = fopen(buf, "w");
	if (!fout) return;
	fprintf(fout, "%d\n", m_ptststats->nD);
	for(int d = 0; d < m_ptststats->nD; ++d) {
		Document &dc = m_ptststats->corpus->docs[d];
		fprintf(fout, "%d %d", m_nK, dc.gndlabel);
		compute_theta(m_ptststats, d);	
		for (int k = 0; k < m_nK; ++k)
			fprintf(fout, " %d:%.10f", k, m_dTheta[k]);
		fprintf(fout, "\n");
	}
	fclose(fout);

	// overall-results
	FILE *fres = fopen(m_params->res_filename, "a");
	fprintf(fres, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) acc: %.4f, train-time: %.4f, test-time: %.4f\n", 
		m_nK, m_dC, m_dNuSqInv, m_params->INITIAL_ALPHA, m_params->INITIAL_BETA,
		(int)m_dDeltaEll, m_params->EM_MAX_ITER, m_params->BURN_IN, m_params->MAX_SAMPLES,
		dAcc, m_dTrainTime, m_dTestTime);
	fflush(fres);
	fclose(fres);

	printf(">>> Saving results & evaluating performace finished! \n");
}

void GibbsMedLDA::save_model_property(Stats* st, Params *param)
{
	char buf[512];	
	sprintf(buf, "%s/model_property.txt", param->dir);
	FILE *file = fopen(buf, "w");
	int d, k, t;
	int nZero = 0;
	compute_phi(st);
	for(k = 0; k < m_nK; ++k)
		for(t = 0; t < m_nV; ++t)
			if (m_dPhi[k][t] < NUMERICAL_ZERO) nZero++;
	double dZeroPercent = double(nZero) / (m_nK * m_nV);
	fprintf(file, "ZeroPercentage: %.10f\n", dZeroPercent);

	double *theta = NULL;
	double dEntropy = 0;
	double entropy = 0;
	double val;

	sprintf(buf, "%s/theta_train.txt", param->dir);
	FILE *thetaFile = fopen(buf, "w");

	for(d = 0; d < st->nD; ++d){
		compute_theta(st, d);
		theta = m_dTheta;
		k = m_nK;
		dEntropy = 0;
		while(k--){
			val = *theta++;
			fprintf(thetaFile, "%.10f ", val);
			if (val > 0)
				dEntropy -= val * log(val);
		}
		fprintf(thetaFile, "\n");
		entropy += dEntropy;
	}
	fclose(thetaFile);
	entropy /= double(st->nD);
	fprintf(file, "AverageEntropy: %.10f\n", entropy);
	fclose(file);
}

void GibbsMedLDA::save_model(const char* model_file)
{
	printf(">>> Saving model ...\n");
	FILE* file = fopen(model_file, "w");
	if (!file) {
		printf("ERROR> Cannot open model file error !");	
		return;
	}

	fprintf(file, "%f\n", m_dC);

	fprintf(file, "%.10f\n", m_dB);
	double* ptr = m_dEta;
	int N = m_nK;
	while(N--)
		fprintf(file, "%5.10f ", *ptr++); 
	fprintf(file, "\n");

	m_ptrnstats->save_wordassigned(file);
	if ( !m_dPhi ) compute_phi( m_ptrnstats );
	for(int i = 0; i < m_nK; ++i){
		for(int j = 0; j < m_nV; ++j){
			fprintf(file, "%.10f ", m_dPhi[i][j]);
		}
		fprintf(file, "\n");
	}
	fclose(file);
	printf(">>> Savning model finished !\n");
}

void GibbsMedLDA::load_model(const char* model_file)
{
	printf(">>> Loading model ...\n");
	FILE* file = fopen(model_file, "r");
	if (!file) {
		printf("ERROR> Cannot open model file error !");	
		return;
	}

	// Loading C
	fscanf(file, "%lf\n", &(m_params->INITIAL_C));
	fprintf(file, "%lf\n", m_dB);

	// Loading eta (svm classifier)
	int N = m_params->NTOPICS;
	m_dEta = new double[N];
	for(int i = 0; i < N; ++i)
		fscanf(file, "%lf", &m_dEta[i]);

	// Loading GibbsMedLDA model
	if (m_ptrnstats) delete m_ptrnstats;
	m_ptrnstats = new Stats();
	m_ptrnstats->load(file, m_params);
	m_nV = m_ptrnstats->nV;

	fclose(file);
	printf(">>> Loading model finished! \n");
}

void GibbsMedLDA::save_lambda(const char* lambda_file)
{
	FILE *fptr = fopen(lambda_file, "w");
	for ( int d=0; d<m_ptrnstats->corpus->num_docs; d++ ) {
		fprintf(fptr, "%.5f\n", m_dLambdaInv[d]);
	}
	fclose(fptr);
}
