#include "stats.h"
#include "cokus.h"
#include "utils.h"
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <vector>
#include <map>
#include <time.h>
using namespace std;

void TopK::reset()
{
	memset(key, 0, sizeof(int) * length);
	memset(value, 0, sizeof(double) * length);
}

void TopK::push_back(int id, double v)
{
	if (v > value[0]) 
	{
		value[0] = v;
		key[0] = id;
		shiftdown(0, length);
	}
}

void TopK::shiftdown(int pos, int rg)
{
	int left = 2 * pos + 1;
	int right = left + 1;
	int min = pos;
	if (left < rg && value[left] < value[pos])
		min = left;		
	if (right < rg && value[right] < value[min])
		min = right;
	
	if (min != pos)
	{
		exchange(min, pos);
		shiftdown(min, rg);
	}
}

void TopK::sort()
{
	int l = length;
	while(l--)
	{
		exchange(0, l);
		shiftdown(0, l);
	}
}

inline void TopK::exchange(int p1, int p2)
{
	int k = key[p1];
	double v = value[p1];
	key[p1] = key[p2];
	value[p1] = value[p2];
	key[p2] = k;
	value[p2] = v;
}

//////////////////////////////////////////////////////////

Stats::Stats()
{
	nD			= 0;
	nV			= 0;
	nK			= 0;
	z			= NULL;
	nw			= NULL;
	nwsum		= NULL;
	nd			= NULL;
	corpus 	= NULL;
}

Stats::~Stats()
{
	Stats::free();
}

void Stats::free()
{
	if (z)
	{
		for (int i = 0; i < nD; ++i)
			if (z[i]) delete[] z[i];
		delete[] z;
	}
	z = NULL;

	if (nw)
	{
		for (int i = 0; i < nK; ++i)
			if (nw[i]) delete[] nw[i];
		delete[] nw;
	}
	nw = NULL;

	if (nwsum) delete[] nwsum;
	nwsum = NULL;

	if (nd)
	{
		for(int i = 0; i < nD; ++i)
			if (nd[i]) delete[] nd[i];
		delete[] nd;
	}
	nd = NULL;

	if (corpus) delete corpus;
	corpus = NULL;

	nD = 0;
	nV	= 0;
	nK	= 0;
}

void Stats::new_stats(Corpus* cps, Params* param)
{
	Stats::free();

	corpus 	= cps;
	nD			= corpus->num_docs;
	nV			= corpus->num_terms;
	nK			= param->NTOPICS;

	z = new int*[nD];
	for(int d = 0; d < nD; ++d)
	{
		Document& dc = corpus->docs[d];
		z[d] = new int[dc.length];
	}

	nwsum = new int[nK];
	memset(nwsum, 0, sizeof(int) * nK);

	nw = new int*[nK];
	for(int k = 0; k < nK; ++k)
	{
		nw[k] = new int[nV];
		memset(nw[k], 0, sizeof(int) * nV);
	}

	nd = new int*[nD];
	for(int d = 0; d < nD; ++d)
	{
		nd[d] = new int[nK];	
		memset(nd[d], 0, sizeof(int) * nK);
	}
}

void Stats::load(const char* filename, Params* param)
{
	FILE* fin = fopen(filename, "r");
	load(fin, param);
	fclose(fin);
}

void Stats::load(FILE* file, Params* param)
{
	Stats::free();

	int length, label, word, topic;
	int nword = 0;

	nK	= param->NTOPICS;

	corpus = new Corpus();
	fscanf(file, "%d\n", &nD);
	corpus->num_docs = nD;
	corpus->docs = new Document[nD];

	z = new int*[nD];
	nd = new int*[nD];
	for(int d = 0; d < nD; ++d)
	{
		nd[d] = new int[nK];	
		memset(nd[d], 0, sizeof(int) * nK);

		Document& dc = corpus->docs[d];
		fscanf(file, "%d %d", &length, &label);
		dc.length	= length;
		dc.gndlabel	= label;
		dc.words		= new int[length];
		z[d]	 		= new int[length];
		for(int n = 0; n < length; ++n)
		{
			fscanf(file, "%d:%d", &word, &topic);
			dc.words[n]		= word;
			z[d][n]			= topic;
			nd[d][topic] 	+= 1;
			if (word >= nword ) nword = word + 1;
		}
	}
	corpus->num_terms = nV = nword;

	nw = new int*[nK];
	for(int k = 0; k < nK; ++k)
	{
		nw[k] = new int[nV];
		memset(nw[k], 0, sizeof(int) * nV);
	}
	nwsum = new int[nK];
	memset(nwsum, 0, sizeof(int) * nK);

	for(int d = 0; d < nD; ++d)
	{
		Document& dc = corpus->docs[d];
		length = dc.length; 
		for(int n = 0; n < length; ++n)
		{
			topic = z[d][n]; 
			nw[topic][dc.words[n]] += 1;
			nwsum[topic] += 1;
		}
	}
}

void Stats::save_wordassigned(const char *filename)
{
	printf("Word-Topic Assignments: %s\n",filename);

	FILE* fout = fopen(filename, "w");
	save_wordassigned(fout);
	fclose(fout);
}

void Stats::save_wordassigned(FILE *file)
{
	if (!file) return;
	fprintf(file, "%d\n", nD);
	for (int d = 0; d < nD; ++d)
	{
		Document& dc = corpus->docs[d];
		fprintf(file, "%d %d", dc.length, dc.gndlabel);	
		for(int n = 0; n < dc.length; ++n)
			fprintf(file, " %d:%d", dc.words[n], z[d][n]);
		fprintf(file, "\n");
	}
}

void Stats::save_topwords(const char* filename, Params* param) 
{
	FILE* fin = fopen(param->mapper_filename, "r");
	if (!fin) return;

	char buf[512];
	map<int, string> mapper;
	int id;
	while( fscanf(fin, "%d %s\n", &id, buf) != EOF )
		mapper.insert(map<int, string>::value_type(id, string(buf)));
	fclose(fin);

	map<int, string>::iterator it;
	FILE* fout = fopen(filename, "w");
	if (!fout) return;

	TopK topk(param->SAVE_TOP_WORDS);
	for(int k = 0; k < nK; ++k)
	{
		topk.reset();
		for(int w = 0; w < nV; ++w)
			topk.push_back(w, (double) nw[k][w] / (double) nwsum[k]);
		topk.sort();
		fprintf(fout, "Topic %d:", k);
		for(int i = 0; i < topk.length; ++i)
		{
			it = mapper.find(topk.key[i]);
			if (mapper.end() != it)
				fprintf(fout, " %s:%f", it->second.c_str(), topk.value[i]);
		}
		fprintf(fout, "\n");
	}
	fclose(fout);
}

void Stats::save_label_only(const char* filename)
{
	FILE* fout = fopen(filename, "w");
	fprintf(fout, "%d\n", nD);
	for(int d = 0; d < nD; ++d)
	{
		Document& dc = corpus->docs[d];
		fprintf(fout, "%d %d\n", dc.gndlabel, dc.predlabel); 
	}
	fclose(fout);
}
